import logging
from typing import List, Optional, Tuple

import pandas as pd
import seaborn as sns

from . import _internals
from ._train_model import train_model

logger = logging.getLogger(f"stream.{__name__}")


def reduce_feature_space(
    setup: _internals.Setup,
    algorithm: str,
    metric: str,
    reference_metric: float,
    acceptable_loss: float,
    hyperparams: Optional[dict] = None,
) -> Tuple[List[str], pd.DataFrame]:
    """
    Reduces the feature space used for model training
    by iteratively removing the feature with the smallest
    impact on model performance.

    Given an algorithm and metric, retrains models with
    one feature removed. Notes the metric when the weakest
    feature is removed and compares it with the acceptable
    loss threshold. If the threshold has not been reached
    the process is repeated.

    If a metric higher than the reference_metric is found,
    it is printed.

    Parameters
    ----------
    setup : Setup
        Dataclass containing the prepared data and further
        configurations.

    algorithm : str
        Algorithm to use for model training.

    metric : str
        Metric to use for evaluating model performance.

    reference_metric : float
        Metric value to use as baseline.

    acceptable_loss : float
        Acceptable loss threshold, e.g. 0.95. Once a value
        of acceptable_loss * reference_metric is undercut,
        the feature space reduction is reduced.

    hyperparams : Optional[dict]
        Hyperparameters to use with the given algorithm.
        Default `None` will use standard hyperparameters.

    Returns
    -------
    Tuple[List[str], pd.DataFrame]
        best_feature_list : List[str]

        metric_feature_df : pd.DataFrame
            Dataframe with columns `metric` and `features` listing
            the metric values achieved with the respective features.
    """
    logger.info("%%% REDUCING FEATURE SPACE")
    # Initiate reference values
    threshold = acceptable_loss * reference_metric
    feature_list = setup.X_train.columns.to_list()
    best_feature_list = feature_list[:]
    metric_feature_df = pd.DataFrame(columns=["metric", "features"])
    new_metric = reference_metric

    logger.info(f"Algorithm selected for feature space reduction: {algorithm}")
    logger.info(f"Metric to optimize for: {metric}")
    logger.info(f"Acceptable loss ratio: {acceptable_loss}")
    logger.info(f"Reference metric: {reference_metric:.3f}")
    logger.info(
        f"Minimum acceptable metric: {acceptable_loss} * {reference_metric:.3f} = {threshold:.3f}"
    )

    # Iteratively remove features
    while (threshold <= new_metric) & (len(feature_list) > 1):
        (worst_feature, new_metric) = _find_worst_feature(
            setup=setup,
            algorithm=algorithm,
            metric=metric,
            feature_list=feature_list,
            hyperparams=hyperparams,
        )
        feature_list.remove(worst_feature)
        metric_feature_df.loc[len(metric_feature_df)] = [new_metric, feature_list[:]]

        # Update reference_metric and threshold if
        # reference_metric was improved upon
        if new_metric > reference_metric:
            reference_metric = new_metric
            threshold = acceptable_loss * reference_metric
            logger.info(
                f"Feature count: {len(feature_list)}, metric: {new_metric:.3f} (new best), removing worst feature: {worst_feature}"
            )
        else:
            logger.info(
                f"Feature count: {len(feature_list)}, metric: {new_metric:.3f}, removing worst feature: {worst_feature}"
            )

        # Update best_feature_list if metric is equal to
        # reference_metric
        if new_metric == reference_metric:
            best_feature_list = feature_list[:]

    # Adding `feature_count`` column to metric_feature_df
    metric_feature_df["feature_count"] = metric_feature_df.features.apply(
        lambda x: len(x)
    )

    logger.info(f"Best {metric} score found: {reference_metric:.3}")
    logger.info(f"Feature list for best score: {best_feature_list}")

    # Plot metric evolution
    sns.lineplot(data=metric_feature_df, x="feature_count", y="metric", sort=True)

    return best_feature_list, metric_feature_df


def _find_worst_feature(
    setup: _internals.Setup,
    algorithm: str,
    metric: str,
    feature_list: List[str],
    hyperparams: Optional[dict] = None,
) -> Tuple[str, float]:
    """
    Finds worst feature given a specific algorithm and
    metric.

    Loops over all features in feature_list, removes the
    individual feature and trains a models instance on the
    remaining features. Afterwards the feature impact is
    evaluated by calculating the achieved metric.

    The feature associated with the largest metric after
    removal of the feature is returned alongside the metric.

    Parameters
    ----------
    setup : Setup
        Dataclass containing the prepared data and further
        configurations.

    algorithm : str
        Algorithm to use for model training.

    metric : str
        Metric to use for evaluating model performance.

    hyperparams : Optional[dict]
        Hyperparameters to use with the given algorithm.
        Default `None` will use standard hyperparameters.

    Returns
    -------
    worst_feature: str
        Name of the feature best to be left out for training.

    new_metric: float
        Metric value achieved when leaving out worst feature.
    """
    # Initiate result list
    removed_feature = []
    new_metric = []

    # Iterate over feature list
    for feature in feature_list:
        _, metrics = train_model(
            algorithm,
            setup,
            feature_list=[feat for feat in feature_list if feat != feature],
            hyperparams=hyperparams,
        )
        removed_feature.append(feature)
        new_metric.append(metrics[metric].values[0])

    # Retrieve worst feature and metric
    new_metric_max = max(new_metric)
    max_index = new_metric.index(new_metric_max)
    worst_feature = removed_feature[max_index]

    return worst_feature, new_metric_max



from typing import List, Optional, Tuple

import pandas as pd
from sklearn.ensemble import ExtraTreesClassifier, RandomForestClassifier
from sklearn.linear_model import (
    LogisticRegression,
    PassiveAggressiveClassifier,
    Perceptron,
    RidgeClassifier,
)
from sklearn.metrics import make_scorer, precision_score
from sklearn.model_selection import cross_validate
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC, LinearSVC
from sklearn.tree import DecisionTreeClassifier, ExtraTreeClassifier

from . import _internals


def train_model(
    algorithm: str,
    setup: _internals.Setup,
    return_models: bool = False,
    feature_list: Optional[List[str]] = None,
    hyperparams: Optional[dict] = None,
) -> Tuple[object, pd.DataFrame]:
    """
    Trains a model instance using specified algorithm and
    calculates various metrics for it.

    Parameters
    ----------
    algorithm : str
        specifies which algorithm to use for training

    setup : Setup
        Dataclass containing the prepared data and further
        configurations.

    return_models: bool = False
        Flag for returning model instances trained on the
        entire training set.

    feature_list: Optional[List[str]] = None
        If provided, will train the model with provided
        feature list. Used within 'reduce_feature_space'
        functionality.

    hyperparams : Optional[dict]
        Hyperparameters to use with the given algorithm.
        Default `None` will use standard hyperparameters.

    Returns
    -------
    model : trained model instance

    metrics : pd.DataFrame
        contains one row listing the different metrics the
        model achieved
    """
    # Instantiate model instance
    model = _internals.get_model_instance(
        algorithm=algorithm,
        trial=None,
        hyperparams=hyperparams,
    )

    # Cross-validate model
    X_train = setup.X_train[feature_list] if feature_list else setup.X_train
    y_train = setup.y_clf_train

    cv_results = cross_validate(
        model,
        X_train,
        y_train,
        scoring={
            "accuracy": "accuracy",
            "precision": make_scorer(
                lambda *args, **kwargs: precision_score(
                    *args, **kwargs, zero_division=0
                )
            ),
            "recall": "recall",
            "f1": "f1",
            "roc_auc": "roc_auc",
        },
        n_jobs=-1,
    )

    # Fit model on full training dataset
    model = model.fit(X_train, y_train) if return_models else None

    # Aggregate metrics
    metrics = _internals.aggregate_metrics(cv_results, algorithm)

    return model, metrics



"""
Contains utilty functionality used by various modules.
"""
import logging
from typing import List, Optional

import pandas as pd
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB

logger = logging.getLogger(f"stream.{__name__}")


def get_available_algos() -> List[str]:
    """
    Returns a list of strings where each string is the
    abbreviation of an algorithm.
    """
    return list(get_algo_dict().keys())


def get_algo_dict() -> dict:
    """
    Returns a dictionary with keys being algorithm abbreviations
    (strings) and values being the callables of the algorithms.
    """
    return {
        "lr": LogisticRegression,
        "nb": GaussianNB,
        # "linearsvc",
        # "rbfsvc",
        # "dt",
        # "extratree",
        # "extratrees",
        # "rf",
        # "ridge",
        # "perceptron",
        # "passive-aggressive",
        # "knn",
    }


def check_include(algo_list: List[str]):
    """Checks `algo_list` for correct strings.

    Parameters
    ----------
    algo_list : List[str]
        List of strings referring to implemented model algorithms.

    Raises
    ------
    LookupError
        In case a string is not matched by an implemented algorithm,
        raises a LookupError.
    """
    available_algos = get_available_algos()
    for entry in algo_list:
        if entry not in available_algos:
            raise LookupError(
                f"'{entry}' was provided in the include parameter, but is not among the avaiable algorithms."
            )


def check_metric(metric: Optional[str]):
    """Checks whether `metric` is among the implemented metrics.
    Raises LookupError if `metric` is not implemented and not None.

    Raises
    ------
    LookupError
        In case `metric` is not matched by an implemented algorithm
        or None, raises a LookupError.
    """
    if metric not in [
        None,
        "algorithm",
        "accuracy",
        "precision",
        "recall",
        "f1",
        "roc_auc",
    ]:
        raise LookupError(
            f"'{metric}' was provided as sort parameter, but is not among the avaiable metrics."
        )


def aggregate_metrics(cv_results: dict, algorithm: str) -> pd.DataFrame:
    """
    Adjusts results of cross validation.

    Parameters
    ----------
    cv_results : dict

    algorithm : str

    Returns
    -------
    metrics : pd.DataFrame
        containing single row with model metrics
    """
    # Round result to 3 digits behind comma
    (accuracy, precision, recall, f1, roc_auc, fit_time,) = (
        round(cv_results[col].mean(), 3)
        for col in [
            "test_accuracy",
            "test_precision",
            "test_recall",
            "test_f1",
            "test_roc_auc",
            "fit_time",
        ]
    )

    return pd.DataFrame(
        {
            "algorithm": [algorithm],
            "accuracy": [accuracy],
            "precision": [precision],
            "recall": [recall],
            "f1": [f1],
            "roc_auc": [roc_auc],
            "Fit time (s)": [fit_time],
        }
    )


def check_feature_scaling(algo_list: List[str], feature_scaling: bool) -> None:
    """Checks for feature_scaling and writes note to logger.INFO if any of
    the utilized algorithms would profit from feature_scaling.

    Parameters
    ----------
    algo_list : List[str]
        List of utilized algorithms.

    feature_scaling : bool
        Option documented in setup object. Specified in config.
    """
    if not feature_scaling:
        affected_algos = [algo for algo in algo_list if algo in ["lr"]]
        if len(affected_algos) > 0:
            logger.warning(
                f"The algorithms {affected_algos} work suboptimally without scaled features. Consider turning it on within the config and rerun pt.modelling.prepare_data()."
            )



import logging
from typing import List, Tuple

import pandas as pd
import yaml
from sklearn.preprocessing import LabelEncoder

from . import _internals

logger = logging.getLogger(f"stream.{__name__}")


def prepare_data(
    train_data: pd.DataFrame,
    config_path: str,
) -> Tuple[_internals.Setup, pd.DataFrame, pd.Series]:
    """Prepares data by
      1) Scaling numeric features
      2) One-Hot-Encoding remaining categorical features
      3) Encoding labels of classification target - if required

    Parameters
    ----------
    train_data : pd.DataFrame
        Dataframe provided by user. Must contain columns specified in config.
    config_path : dict
        Path to config containing input features and preparational options.

    Returns
    -------
    Tuple
        Setup, which will be handed to function compare_algorithms
        for further processing.
        X_sample & y_sample showing what the data looks like after
        preparation.
    """
    logger.info("%%% PREPARING DATA")
    # Loading config
    config = _init_config(path=config_path)

    # Checking input
    _check_clf_target(train_data, config["modelling"]["target_clf"])
    _check_numeric_cols(train_data, config["modelling"]["numeric_cols"])
    _check_categorical_cols(train_data, config["modelling"]["categorical_cols"])

    # Get features and targets from config
    original_features = _get_original_features(config)
    target_clf = config["modelling"]["target_clf"]

    ###################
    ### Preparation ###

    # Assembling preprocessing pipeline
    prep_pipe = _internals.get_prep_pipeline(config=config)

    # Composing X_train
    X_train = prep_pipe.fit_transform(train_data[original_features])
    logger.info("Applied preprocessing transformations")

    # Encoding target_clf
    y_train, y_clf_encoder = _encode_y_train(train_data[target_clf])

    return (
        _internals.Setup(
            X_train=X_train,
            y_clf_train=y_train,
            y_clf_encoder=y_clf_encoder,
            feature_scaling=config["modelling"]["feature_scaling"],
            prep_pipe=prep_pipe,
        ),
        X_train.head(),
        y_train.head(),
    )


def _init_config(path: str):
    with open(path, "r") as stream:
        try:
            config = yaml.safe_load(stream)
            logger.info(f"Read {path}: \n {config}")
            return config
        except yaml.YAMLError as exc:
            print(exc)


def _check_clf_target(train_data: pd.DataFrame, clf_col: str = None) -> None:
    """Raises LookupError if clf target column not in
    train_data dataframe.

    Parameters
    ----------
    train_data : pd.DataFrame
        Lookuperror

    clf_col : str
    """
    if clf_col:
        if clf_col not in train_data:
            raise LookupError(
                f"{clf_col}, which was provided as 'target_clf', is not in\
                    train_data dataframe. Check existence and spelling."
            )


def _check_numeric_cols(
    train_data: pd.DataFrame, numeric_cols: List[str] = None
) -> None:
    """Raises LookupError if one or more of the numeric
    columns listed in config are missing.

    Parameters
    ----------
    train_data : pd.DataFrame

    numeric_cols : List[str]
    """
    if numeric_cols:
        for col in numeric_cols:
            if col not in train_data:
                raise LookupError(
                    f"{col}, which was provided in 'numeric_cols', is not in\
                        train_data dataframe. Check existence and spelling."
                )


def _check_categorical_cols(
    train_data: pd.DataFrame, categorical_cols: List[str]
) -> None:
    """Raises LookupError if one or more of the categorical
    columns listed in config are missing.

    Parameters
    ----------
    train_data : pd.DataFrame

    categorical_cols : List[str]
    """
    if categorical_cols:
        for col in categorical_cols:
            if col not in train_data:
                raise LookupError(
                    f"{col}, which was provided in 'categorical_cols', is not in\
                        train_data dataframe. Check existence and spelling."
                )


def _encode_y_train(y: pd.Series) -> Tuple[pd.Series, LabelEncoder]:
    """Encodes target column of classification problems if required.
    Regardless of encoding, downcasts target column to save memory.

    Parameters
    ----------
    y : pd.Series
        Column to encode

    Returns
    -------
    Tuple[pd.Series, LabelEncoder]
        Encoded column and fitted LabelEncoder containing class labels.
    """
    if not _check_encoding_necessity(y):
        return pd.to_numeric(y, downcast="integer"), LabelEncoder()
    else:
        le = LabelEncoder().fit(y)
        y_trans = pd.Series(le.transform(y), name=y.name)
        logger.info(
            f"Encoded target variable using classes: {[(i, class_) for i, class_ in enumerate(le.classes_)]}"
        )
        return pd.to_numeric(y_trans, downcast="integer"), le


def _check_encoding_necessity(y: pd.Series) -> bool:
    """Checks column for dtype integer.

    Parameters
    ----------
    y : pd.Series
        Target column of classification problem.

    Returns
    -------
    bool
        Boolean flag used for encoding
    """
    return True if (y.dtype == "O") else False


def _get_original_features(config: dict) -> List[str]:
    """Returns list of features passed to prepare_data function.

    Parameters
    ----------
    config : dict

    Returns
    -------
    List[str]
        Combination of numeric, categorical and date features listed
        in config.yml.
    """
    numeric_features = (
        config["modelling"]["numeric_cols"]
        if config["modelling"]["numeric_cols"]
        else []
    )
    categorical_features = (
        config["modelling"]["categorical_cols"]
        if config["modelling"]["categorical_cols"]
        else []
    )

    return [*numeric_features, *categorical_features]



from ._models import get_model_instance
from ._prep_pipeline import OHE, get_prep_pipeline
from ._setup import Setup
from ._utils import (
    aggregate_metrics,
    check_feature_scaling,
    check_include,
    check_metric,
    get_algo_dict,
    get_available_algos,
)



"""
Logic related to exporting a trained model with a corresponding
preprocessing pipeline.
"""
import abc
import datetime
import logging
from copy import deepcopy
from pathlib import Path
from typing import Optional, Union

import joblib
import numpy as np
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import LabelEncoder

from . import _internals

logger = logging.getLogger(f"stream.{__name__}")


def export_model(setup: _internals.Setup, model, target_dir: str) -> None:
    """Exports provided `model` with preprocessing pipeline stored in
    setup.

    The output file is binarized using joblib and saved to `target_dir`.
    The output filename follows the syntax
      `model_{date}_{algo}_#{number}.joblib`
    where
    - `date` is the current date in format YYYY-MM-DD
    - `algo` is the utilised model algorithm, and
    - `number` is a number assigned when other models with the same
    name already exist.

    Example
    -------
    >>> pt.modelling.export_model(
            setup=setup,
            model=model_list[0],
            target_dir="./",
        )

    Parameters
    ----------
    setup : _internals.Setup
        Setup object containing preprocessing pipeline.

    model :
        Trained model instance.

    target_dir : str
        Target directory for saving the output file to.
    """
    full_pipe = _combine_pipeline_and_model(
        prep_pipe=setup.prep_pipe, model=model, y_clf_encoder=setup.y_clf_encoder
    )
    export_name = _get_export_name(target_dir=target_dir, model=model)
    joblib.dump(full_pipe, Path(target_dir) / export_name)
    logger.info(f"Exported modelling pipeline to '{Path(target_dir) / export_name}'")


def _combine_pipeline_and_model(
    prep_pipe: Pipeline, model, y_clf_encoder: LabelEncoder
) -> Pipeline:
    """Adds `model` as last step to `prep_pipe`."""
    # Making deep copy, so that the model-step is not added to the
    # original pipeline object.
    prep_pipe_copy = deepcopy(prep_pipe)
    prep_pipe_copy.steps.append(
        ("model_wrapper", ModelWrapper(estimator=model, y_clf_encoder=y_clf_encoder))
    )
    # prep_pipe_copy.steps.append(("model", model))
    return prep_pipe_copy


def _get_export_name(target_dir: str, model) -> str:
    """Generates filename for `full_pipe`."""
    date = str(datetime.date.today())
    algo = _get_algo_abbreviation(model_type=type(model))
    temp_name = f"model_{date}_{algo}_#"
    new_number = _get_new_number(target_dir=target_dir, temp_name=temp_name)
    return (
        f"model_{date}_{algo}_#{new_number}.joblib"
        if new_number
        else f"model_{date}_{algo}_#1.joblib"
    )


def _get_algo_abbreviation(model_type: abc.ABCMeta) -> str:
    """Returns string abbreviation for provided algorithm."""
    algo_dict = _internals.get_algo_dict()
    return [key for key in algo_dict.keys() if algo_dict[key] == model_type][0]


def _get_new_number(target_dir: str, temp_name: str) -> Optional[int]:
    """Determines the number value for the new file name by
    checking `target_dir` for pre-existing files with matching
    name pattern.

    Parameters
    ----------
    target_dir : str
        Target directory for new file. This is also where for
        pre-existing files of similar name is checked.

    temp_name : str
        Basic name structure for new file.

    Returns
    -------
    Optional[int]
        Integer value of the new number, or None.
        None is returned when no pre-existing file is found.
    """
    files_in_target_dir = [
        str(path.name) for path in Path(target_dir).glob("./*") if path.is_file()
    ]
    files_matching_pattern = [
        filename for filename in files_in_target_dir if temp_name in filename
    ]
    if len(files_matching_pattern) > 0:
        last_entry = sorted(files_matching_pattern)[-1]
        last_number = int(last_entry.rsplit("#")[-1].split(".joblib")[0])
        return last_number + 1
    else:
        return None


class ModelWrapper:
    """Wrapper class for estimator and label encoder.

    In case of a classification task the target variable is often
    encoded as string format. During preprocessing such target columns
    are encoded to numerical values so that the subsequent estimator can
    handle the data. This wrapper class facilitates the decoding of the
    estimator's prediction into the orignal format.

    Attributes
    ----------
    estimator : Estimator in scikit-learn format
        Estimator trained during a previous step.
    y_clf_encoder : LabelEncoder
        Retrieved from Setup object, i.e. setup.y_clf_encoder

    Methods
    -------
    predict(X)
        Calls the predict()-method of the estimator and applies decoding
        of the label encoder if one was used.
    predict_proba(X)
        Calls the estimator's predict_proba()-method.
    """

    def __init__(self, estimator, y_clf_encoder: Union[LabelEncoder, None]) -> None:
        self.estimator = estimator
        self.y_clf_encoder = y_clf_encoder

    def predict(self, X) -> np.ndarray:
        if self.y_clf_encoder:
            return self.y_clf_encoder.inverse_transform(self.estimator.predict(X))
        else:
            return self.estimator.predict(X)

    def predict_proba(self, X) -> np.ndarray:
        return self.estimator.predict_proba(X)



from dataclasses import dataclass

import pandas as pd
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import LabelEncoder


@dataclass
class Setup:
    """
    Container class to be returned when using
    modelling.prepare_data(). It holds all the information
    acquired during this preparation step.

    Attributes
    ----------
    X_train : pd.DataFrame
        Training dataset without target.

    y_clf_train : pd.Series
        Classifier target of training dataset.

    y_clf_encoder : LabelEncoder
        LabelEncoder used for encoding classification target.

    feature_scaling : bool
        Configuration flag indicating whether prepare_data performed
        feature_scaling.

    prep_pipe : Pipeline
        Preprocessing pipeline used before model fitting/prediction.
    """

    X_train: pd.DataFrame
    y_clf_train: pd.Series
    y_clf_encoder: LabelEncoder
    feature_scaling: bool
    prep_pipe: Pipeline



from ._compare_algorithms import compare_algorithms
from ._export_model import export_model
from ._prepare_data import prepare_data
from ._reduce_feature_space import reduce_feature_space
from ._tune_hyperparams import tune_hyperparams



"""
Container for implemented models.

Each implemented model is described by a class containing attributes with
  1) the base model to use when comparing algorithms
  2) the hyperparameter space to search when tuning

When a model is required, it can be requested using the function
`_get_base_model(type="classification", "lr")`.
"""

from abc import ABC, abstractmethod
from typing import Optional

import optuna
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB


def get_model_instance(
    algorithm: str,
    trial: Optional[optuna.Trial] = None,
    hyperparams: Optional[dict] = None,
):
    """Returns model instance for `algorithm`.

    When a `trial` instance is provided, the hyperparams of the
    model instance are set using optuna's trial object.

    If `hyperparams` are provided, the algorithm is configured with
    these parameters.

    A simultaneous occurence of `trial` and `hyperparams` is not
    accounted for and will raise an error.

    Parameters
    ----------
    algorithm : str
        Algorithm specifying what type of model to return.

    trial : Optional[optuna.Trial]
        Optuna's trial object used to specify hyperparameters to use.

    hyperparams : Optional[dict]
        Hyperparameters to use with the given algorithm.
        Default `None` will use standard hyperparameters.

    Returns
    -------
    model_instance
        A model instance based on `algorithm` with base or adjusted
        hyperparameters depending on a trial-object being provided.
    """
    _check_correct_params(trial, hyperparams)

    if algorithm == "lr":
        algo = ClfLinearRegression
    elif algorithm == "nb":
        algo = ClfGaussianNB
    else:
        raise LookupError(f"'{algorithm}' is not among the avaiable algorithms.")

    return (
        algo(trial=trial).get_model().set_params(**hyperparams)
        if hyperparams
        else algo(trial=trial).get_model()
    )


def _check_correct_params(
    trial: Optional[optuna.Trial] = None,
    hyperparams: Optional[dict] = None,
) -> None:
    """Raises an error if both `trial` and `hyperparams` are not `None`."""
    if (trial is not None) & (hyperparams is not None):
        raise ValueError(
            """A model instance was to be instantiated with specific hyperparameters
               and tunable hyperparameters. This combination should not be reached.
               Please reach out to a maintainer.
        """
        )


class ModelContainer(ABC):
    """
    Abstract class for implementing algorithms into pycarrot.

    Each child class must implement two attributes:
      1) self.base_model
      2) self.hyperparams

    `base_model` is a model instance with default hyperparameters.
    "Default" in this context typically refers to the library's standard
    hyperparameters, sometimes with slight adjustments for the target audience
    of pycarrot, e.g. max_iter for LinearRegression and larger datasets.

    `tuning_params` is either None or a dictionary if the child class is
    instaniated with a trial object. Keys of that dictionary must mirrow
    the model's hyperparameter names and the values consist of optuna's
    trial suggestions. See example below.

    This parent class also provides a generic get_model function that
    returns a model instance based on whether a trial object has been
    provided or not.

    Example
    -------
    >>> self.base_model = LogisticRegression(
            solver="saga", max_iter=1000, penalty="elasticnet", l1_ratio=1
        )
    >>> self.tuning_params = (
            {
                "C": trial.suggest_float("C", 1e-6, 1e2, log=True),
                "penalty": trial.suggest_categorical(
                    "penalty", ["elasticnet", "none"]
                ),
                "l1_ratio": trial.suggest_uniform("l1_ratio", 0, 1),
            }
            if trial
            else None
        )
    """

    @abstractmethod
    def __init__(self, trial: Optional[optuna.Trial] = None):
        self.base_model = None
        self.tuning_params = None

    def get_model(self):
        """Returns model instance based on child class attributes."""
        return (
            self.base_model
            if self.tuning_params is None
            else self.base_model.set_params(**self.tuning_params)
        )


class ClfLinearRegression(ModelContainer):
    """
    ModelContainer for classification algorithm `LinearRegression`.

    See: https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LogisticRegression.html

    Hyperparameter choice:
      "solver": Algorithm for optimization set to `saga` as it supports
         all penalty norms and is typically fastest on large datasets.
         Requires normalized features for good convergence. Typically
         only minor differences in model quality are observed by varying
         the solver.
      "penalty": Norm of the penalty. Typically strong impact. Only includes
         `elasticnet` as `l1` and `l2` are special cases covered by l1_ratio.
         Forgoing the penalty (penalty=`none`) is covered by very large values
         of `C`.
      "C": Inverse of regularization strength. In general strong impact.
      "l1_ratio": If solver `elasticnet` is used, the l1_ratio determines
         the combination ratio between l1 and l2. 0 corresponds to pure `l1`
         while a value of 1 results in pure `l2`.
    """

    def __init__(self, trial: Optional[optuna.Trial] = None):
        self.base_model = LogisticRegression(
            solver="saga", max_iter=1000, penalty="elasticnet", l1_ratio=1
        )
        self.tuning_params = (
            {
                "C": trial.suggest_float("C", 1e-6, 1e2, log=True),
                "l1_ratio": trial.suggest_uniform("l1_ratio", 0, 1),
            }
            if trial
            else None
        )


class ClfGaussianNB(ModelContainer):
    """
    ModelContainer for classification algorithm `GaussianNB`.

    See: https://scikit-learn.org/stable/modules/generated/sklearn.naive_bayes.GaussianNB.html

    Hyperparameter choice:
      "var_smoothing": Portion of the largest variance of all features
      that is added to variances for calculation stability. Only tunable
      feature.
    """

    def __init__(self, trial: Optional[optuna.Trial] = None):
        self.base_model = GaussianNB()
        self.tuning_params = (
            {
                "var_smoothing": trial.suggest_float(
                    "var_smoothing", 1e-12, 1e0, log=True
                ),
            }
            if trial
            else None
        )



import pandas as pd


def get_distribution(df: pd.DataFrame):
    for i, col in enumerate(df):
        print(f"COLUMN: {col}")
        print("-----------------------------")
        print(f"Dtype: {df[col].dtype}")
        print(f"Unique values: {df[col].nunique()}")
        print(f"np.nan count: {df[col].isna().sum()}")
        print(
            f"Most common values: {df[col].value_counts(ascending=False).head().index.values}"
        )
        print()



"""
Contains tools for developing pycarrot.
"""
from typing import Tuple

import numpy as np
import pandas as pd

from .modelling import _prepare_data


def load_dataset(type_: str, name: str) -> Tuple[pd.DataFrame, dict]:
    """Loads specified dataset as a pandas dataframe and a
    pre-generated config file to be used with pycarrot.

    Parameters
    ----------
    type : str
        Specification of type of dataset. Either clf or reg
        for classification and regression respectively.
    name : str
        Identifier for the dataset. Possible values:
        Classification:
            1) numeric_noNAs_binary - based on breast_cancer
               dataset
        Regression:
        -

    Returns
    -------
    Tuple[pd.DataFrame, dict]
        Dataframe and config dictionary for processing with
        pycarrot.
    """
    if type_ == "clf":
        path_stem = f"./dev_datasets/classification/{name}"
    elif type_ == "reg":
        path_stem = f"./dev_datasets/regression/{name}"

    df = pd.read_parquet(f"{path_stem}.parquet")
    config = _prepare_data._init_config(path=f"{path_stem}.yml")
    return df, config


def drop_random_entries(
    df: pd.DataFrame, pct: float = 0.1, keep: str = None
) -> pd.DataFrame:
    """Drops entries from pd.DataFrame at random. Used for creating
    dataframes with missing information.

    Parameters
    ----------
    df : pd.DataFrame
        Original dataframe without missing values.
    pct : float, optional
        Fraction of values to drop, by default 0.1
    keep : str, optional
        Column name of column to leave unmodified (typically target variable),
        by default None

    Returns
    -------
    pd.DataFrame
        Dataframe with randomly dropped values.
    """
    df = df.copy()
    features = [col for col in df.columns if col != keep]
    num_entries_to_drop = int(np.floor(len(features) * len(df) * pct))
    for _ in range(num_entries_to_drop):
        col = np.random.choice(features)
        loc = np.random.randint(len(df))
        df.loc[loc, col] = np.nan
    return df



import logging

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder, StandardScaler

logger = logging.getLogger(f"stream.{__name__}")


def get_prep_pipeline(config: dict) -> Pipeline:
    """Return unfitted preprocessing pipeline.

    Preprocessing steps include:
      1) Standardization of numeric columns
      2) One-Hot-Encoding of 'object' columns

    Parameters
    ----------
    config : Dict
        Configuration read from provided config.yml

    Returns
    -------
    Pipeline
        Unfitted preprocessing pipeline.
    """
    steps = [
        # StandardScaler (optional)
        (
            "scaler",
            CustomStandardScaler(numeric_cols=config["modelling"]["numeric_cols"]),
        )
        if config["modelling"]["feature_scaling"]
        else "skip",
        # OneHotEncoder
        ("ohe", OHE()),
    ]

    prep_pipe = Pipeline(steps=[step for step in steps if step != "skip"])
    step_names = [step[0] for step in prep_pipe.steps]
    logger.info(f"Created preprocessing pipeline with following steps: {step_names}")
    return prep_pipe


class CustomStandardScaler(BaseEstimator, TransformerMixin):
    """Transformer class applying scaling to numeric features."""

    def __init__(self, numeric_cols):
        self.scaler = StandardScaler()
        self.numeric_cols = numeric_cols

    def fit(self, X, y=None):
        self.scaler.fit(X.loc[:, self.numeric_cols])
        return self

    def transform(self, X, y=None):
        X.loc[:, self.numeric_cols] = self.scaler.transform(X[self.numeric_cols])
        return X


class OHE(BaseEstimator, TransformerMixin):
    """Transformer class performing One-Hot-Encoding of 'object' typed columns."""

    def __init__(self):
        self.ohe = OneHotEncoder(handle_unknown="ignore")
        self.col_names = None

    def fit(self, X, y=None):
        categorical_data = X.select_dtypes(include=("object"))
        if len(categorical_data.columns) > 0:
            self.ohe.fit(categorical_data)
            self.col_names = self.ohe.get_feature_names_out(categorical_data.columns)
        return self

    def transform(self, X, y=None):
        # Checking whether encoder was fitted
        if self.col_names is not None:
            non_categorical_data = X.select_dtypes(exclude=("object"))
            categorical_data = X.select_dtypes(include=("object"))
            ohe_data = self.ohe.transform(categorical_data).toarray()
            ohe_df = pd.DataFrame(
                data=ohe_data, columns=self.col_names, index=non_categorical_data.index
            ).astype("uint8")
            return pd.concat([non_categorical_data, ohe_df], axis=1)
        else:
            return X



"""
Contains logic for hyperparameter tuning.

Example
-------
>>> import pycarrot as pc
>>> from sklearn.datasets import load_breast_cancer

>>> bc = load_breast_cancer()
>>> X = pd.DataFrame(bc.data, columns=bc.feature_names)
>>> y = pd.Series(bc.target, name="class")
>>> df = pd.concat([X, y], axis=1)

>>> config = pt.init_config("./config_bc.yml")

>>> setup, X_sample, y_sample = pt.modelling.prepare_data(
        train_data=df,
        config=config,
    )

>>> compare_df, model_list, opt_history_dict = (
            pt.modelling.tune_hyperparams(
                   setup=setup,
                   include=["lr", "knn"],
                   optimize="f1",
                   n_trials=50,
                   return_models=True,
            )
    )
"""
import logging
from typing import List, Optional, Tuple

import numpy as np
import optuna
import pandas as pd
from sklearn.metrics import make_scorer, precision_score
from sklearn.model_selection import cross_validate

from . import _internals

logger = logging.getLogger(f"stream.{__name__}")

# Turning off logging information during tuning
optuna.logging.set_verbosity(optuna.logging.WARNING)


def tune_hyperparams(
    setup: _internals.Setup,
    include: List[str],
    optimize: str,
    n_trials: int = 20,
    feature_list: Optional[List[str]] = None,
    return_models: bool = False,
) -> Tuple[pd.DataFrame, list, dict]:
    """Tunes the algorithms provided in the `include` parameter.

    Example
    -------
    >>> compare_df, model_list = (
                pt.modelling.tune_hyperparams(
                    setup=setup,
                    include=["lr", "knn"],
                    optimize="f1",
                    n_trials=50,
                    return_models=True,
                )
        )

    Parameters
    ----------
    setup : Setup
        Dataclass containing the prepared data and further
        configurations.

    include : List[str]
        List of identifier strings for algorithms to tune.

    optimize : str
        String identifier of metric to optimize for.

    n_trials : int, optional
        Number of hyperparameter combinations to evaluate,
        by default 20

    feature_list: Optional[List[str]] = None
        If provided, will tune the model with provided
        feature list. Useful after reduce_feature_space()
        has been evaluated.

    return_models : bool, optional
        Flag for training models on the entire dateset using the
        best hyperparameter combinations found, by default False

    Returns
    -------
    Tuple[pd.DataFrame, list, dict]
        compare_df : pd.DataFrame
            sorted overview of algorithm performance

        model_list : list
            List of model instances trained on entire training data
            using best hyperparameter combination found. List is ordered
            by achieved metric. Only runs if return_models == True.
            Otherwise returns list of None.

        opt_history_dict : dict
            Dictionary containing algorithms as keys and figures of the
            optimization history as values.
            Use `opt_history_dict["algorithm"].show()` in a jupyter notebook
            to plot the graph.
    """
    logger.info("%%% TUNING HYPERPARAMETERS")

    # Checking inputs
    _internals.check_include(algo_list=include)
    _internals.check_metric(metric=optimize)
    _internals.check_feature_scaling(
        algo_list=include, feature_scaling=setup.feature_scaling
    )

    # Preparing empty compare_df and model_dict
    # with populating occuring later
    compare_df = pd.DataFrame(
        {},
        columns=[
            "algorithm",
            "accuracy",
            "precision",
            "recall",
            "f1",
            "roc_auc",
            "Fit time (s)",
            "hyperparams",
        ],
    )
    model_dict = {}
    opt_history_dict = {}

    logger.info(f"Algorithms: {include}")
    logger.info(f"Metric: {optimize}")
    logger.info(f"Trials per algorithm: {n_trials}")

    for algorithm in include:
        # Preparation of tuning
        objective = _get_objective_function(
            algorithm=algorithm,
            optimize=optimize,
            X=setup.X_train[feature_list] if feature_list else setup.X_train,
            y=setup.y_clf_train,
        )
        study = optuna.create_study(
            study_name=f"study_{algorithm}", direction="maximize"
        )

        # Tuning
        logging_callback = TrialLoggingCallback()
        study.optimize(objective, n_trials=n_trials, callbacks=[logging_callback])

        # Aggreration of results
        metrics, hyperparams = _get_best_result(study)
        compare_df.loc[len(compare_df)] = [
            *metrics.values[0],
            hyperparams,
        ]
        opt_history_dict[algorithm] = optuna.visualization.plot_optimization_history(
            study
        )

        # Training model on entire dataset
        if return_models:
            model_dict[algorithm] = (
                _internals.get_model_instance(algorithm=algorithm)
                .set_params(**hyperparams)
                .fit(setup.X_train, setup.y_clf_train)
            )

    compare_df = compare_df.sort_values(by=optimize, ascending=False).reset_index(
        drop=True
    )
    model_list = (
        [model_dict[key] for key in compare_df["algorithm"]] if return_models else []
    )
    logger.info(f"Tuning summary:\n {compare_df}")

    return compare_df, model_list, opt_history_dict


def _get_objective_function(
    algorithm: str, optimize: str, X: pd.DataFrame, y: pd.Series
) -> object:
    """Returns model specific objective function for usage in optuna's
    stduy.optimize() function.

    Parameters
    ----------
    algorithm : str
        Model algorithm to use in objective function.

    optimize : str
        Metric to optimize for.

    X : pd.DataFrame
        Dataframe containing training features.

    y : pd.Series
        Series containing target variable.

    Returns
    -------
    object
        `objective function` object
    """

    def objective_fn(trial: optuna.Trial) -> np.float64:
        """Objective function for usage within optuna's study.optimize().

        Parameters
        ----------
        trial : optuna.Trial
            Handled automatically by study.optimize()

        Returns
        -------
        np.float64
            Mean metric of all folds.
        """
        model = _internals.get_model_instance(algorithm=algorithm, trial=trial)
        trial.set_user_attr("hyperparams", model.get_params())
        cv_results = cross_validate(
            model,
            X,
            y,
            scoring={
                "accuracy": "accuracy",
                "precision": make_scorer(
                    lambda *args, **kwargs: precision_score(
                        *args, **kwargs, zero_division=0
                    )
                ),
                "recall": "recall",
                "f1": "f1",
                "roc_auc": "roc_auc",
            },
            n_jobs=-1,
        )
        metrics = _internals.aggregate_metrics(cv_results, algorithm)
        trial.set_user_attr("metrics", metrics)

        return cv_results[f"test_{optimize}"].mean()

    return objective_fn


def _get_best_result(study: optuna.Study) -> Tuple[float, dict]:
    """Returns metrics and hyperparameters of best model found
    during study run.

    Parameters
    ----------
    study : optuna.Study

    Returns
    -------
    Tuple[float, dict]
        metrics : pd.DataFrame
        hyperparams : dict
    """
    return (
        study.best_trial.user_attrs["metrics"],
        study.best_trial.user_attrs["hyperparams"],
    )


class TrialLoggingCallback:
    def __call__(
        self, study: optuna.study.Study, trial: optuna.trial.FrozenTrial
    ) -> None:
        logger.info(
            f"Trial {trial.number} finished with value: {study.trials[trial.number].value:.4f} and parameters: {trial.params}. "
            f"Best is trial {study.best_trial.number} with value: {study.best_value:.4f}"
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pytelligence', 'pytelligence.modelling', 'pytelligence.modelling._internals']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'optuna>=2.10.1,<3.0.0',
 'pandas>=1.4.3,<2.0.0',
 'plotly>=5.10.0,<6.0.0',
 'scikit-learn>=1.1.1,<2.0.0',
 'seaborn>=0.11.2,<0.12.0']

setup_kwargs = {
    'name': 'pytelligence',
    'version': '0.2.2',
    'description': 'pycaret clone aimed for simplicity and production ready code',
    'long_description': None,
    'author': 'Friedrich G. Froebel',
    'author_email': 'froebel.business@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import logging
from typing import List, Optional, Tuple

import pandas as pd

from . import _internals
from ._train_model import train_model

logger = logging.getLogger(f"stream.{__name__}")


def compare_algorithms(
    setup: dict,
    include: List[str] = _internals.get_available_algos(),
    sort: Optional[str] = None,
    return_models: bool = False,
    feature_list: Optional[List[str]] = None,
) -> Tuple[pd.DataFrame, List, List]:
    """Calculates various metrics for different machine learning
    algorithms.

    Parameters
    ----------
    setup : dict stemming from pt.modelling.prepare_data(...)

    include : optional list of strings
        declares what algorithms to compare

    sort : optional str
        Defines how compare_df is sorted. Possible values are accuracy,
        precision, recall, f1 & roc_auc.

    return_models: bool
        Flag for returning model instances trained on the
        entire training set. By default set to false to
        save on computational time.

    feature_list: Optional[List[str]] = None
        If provided, will compare the algorithms with provided
        feature list. Useful after reduce_feature_space()
        has been evaluated.

    Returns
    -------
    compare_df : pd.DataFrame
        sorted overview of algorithm performance

    algo_list : list
        List of algorithms ordered by sort metric.

    model_list : list
        Trained model instance if return_models == True.
        Otherwise returns list of None.
    """
    logger.info("%%% COMPARING ALGORITHMS")

    # Checking inputs
    _internals.check_include(include)
    _internals.check_metric(metric=sort)
    _internals.check_feature_scaling(
        algo_list=include, feature_scaling=setup.feature_scaling
    )

    # Preparing empty compare_df and model_dict
    # with populating occuring later
    compare_df = _prepare_compare_df()
    model_dict = {}

    # Training models
    for algorithm in include:
        logger.info(f"Evaluating {algorithm}...")
        model, metrics = train_model(
            algorithm=algorithm,
            setup=setup,
            return_models=return_models,
            feature_list=feature_list,
        )
        compare_df = pd.concat([compare_df, metrics])
        model_dict[algorithm] = model

    # Sort compare_df
    if sort:
        compare_df = compare_df.sort_values(
            by=[sort, "Fit time (s)"],
            ascending=[False, True],
        ).reset_index(drop=True)

    algo_list = compare_df["algorithm"].to_list()
    model_list = [model_dict[key] for key in compare_df["algorithm"]]

    return compare_df, algo_list, model_list


def _prepare_compare_df() -> pd.DataFrame:
    """
    Creates compare_df dataframe with required columns,
    but no entries yet.

    Returns
    -------
    compare_df : pd.DataFrame
    """
    return pd.DataFrame(
        columns=[
            "algorithm",
            "accuracy",
            "precision",
            "recall",
            "f1",
            "roc_auc",
            "Fit time (s)",
        ]
    )



import datetime
import logging
from pathlib import Path

import colorlog

from . import dev_tools, feat_analysis, modelling

# Create logs folder if it does not exists
if not Path("./logs").exists():
    Path("./logs").mkdir()
    print(f"Created directory for logs at '{Path('./logs').resolve()}'")

# Set up root logger, and add a file handler to root logger
logging.basicConfig(
    filename=f"./logs/log_{datetime.datetime.now().strftime('%Y-%m-%d_%H:%M')}.log",
    level=logging.DEBUG,
    format="[%(levelname)1.1s %(asctime)s] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

# Create logger for handling stream
shandler = colorlog.StreamHandler()
shandler.setFormatter(
    colorlog.ColoredFormatter(
        "%(log_color)s[%(levelname)1.1s %(asctime)s]%(reset)s %(message)s",
        "%Y-%m-%d %H:%M:%S",
    )
)
stream_logger = colorlog.getLogger("stream")
stream_logger.addHandler(shandler)
